############################################
# imports
############################################

import pandas as pd
from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier
from fgclustering.forest_guided_clustering import *


############################################
# Tests
############################################


def test_forest_guided_clustering():
    # parameters
    max_K = 6
    method_clustering = "pam"
    init_clustering = "random"
    max_iter_clustering = 100
    discart_value_JI = 0.6
    bootstraps_JI = 100
    n_jobs = 3

    # test data
    X, y = make_classification(
        n_samples=300,
        n_features=10,
        n_informative=4,
        n_redundant=1,
        n_classes=2,
        n_clusters_per_class=1,
        random_state=42,
    )
    X = pd.DataFrame(X)

    model = RandomForestClassifier(
        max_depth=10,
        max_features="sqrt",
        max_samples=0.8,
        bootstrap=True,
        oob_score=True,
        random_state=42,
    )
    model.fit(X, y)

    data_classification = X
    data_classification["target"] = y

    # initialize and run fgclustering object
    fgc = FgClustering(model=model, data=data_classification, target_column="target")
    fgc.run(
        max_K=max_K,
        method_clustering=method_clustering,
        init_clustering=init_clustering,
        max_iter_clustering=max_iter_clustering,
        discart_value_JI=discart_value_JI,
        bootstraps_JI=bootstraps_JI,
        n_jobs=n_jobs,
    )

    # obtain optimal number of clusters
    assert fgc.k == 2, "error: wrong optimal k calculated"



############################################
# imports
############################################

import numpy as np
from tqdm import tqdm

import kmedoids
from joblib import Parallel, delayed
import collections, functools, operator
from numba import njit, prange

import fgclustering.statistics as statistics

# import warnings
# warnings.filterwarnings('ignore')

############################################
# Optimize number of clusters k
############################################


def _compute_jaccard_matrix(clusters, indices_bootstrap_clusters, indices_original_clusters):
    """Compute Jaccard Index between all possible cluster combinations of original vs bootstrapped clustering.

    :param clusters: Clustering labels.
    :type clusters: numpy.ndarray
    :param indices_bootstrap_clusters: Dictionary with cluster labels as keys and index of instances that
        belong to the respective cluster as labels for boostrapped clustering.
    :type indices_bootstrap_clusters: dict
    :param indices_original_clusters: Dictionary with cluster labels as keys and index of instances that
        belong to the respective cluster as labels for original clustering.
    :type indices_original_clusters: dict
    :return: Jaccard Index for all cluster combinations.
    :rtype: numpy.ndarray
    """
    indices_bootstrap_all = np.unique(
        [
            index
            for i, cluster_bootstrap in enumerate(clusters)
            for index in indices_bootstrap_clusters[cluster_bootstrap]
        ]
    )
    jaccard_matrix = np.zeros([len(clusters), len(clusters)])

    for i, cluster_original in enumerate(clusters):
        for j, cluster_bootstrap in enumerate(clusters):
            indices_bootstrap = indices_bootstrap_clusters[cluster_bootstrap]
            indices_original = indices_original_clusters[cluster_original]

            # only compute overlap for instances that were in the whole bootstrap sample
            indices_original = indices_original.intersection(indices_bootstrap_all)

            intersection = indices_original.intersection(indices_bootstrap)
            union = indices_original.union(indices_bootstrap)

            jaccard_matrix[i, j] = len(intersection) / len(union)

    return jaccard_matrix


@njit
def _get_bootstrap(M, bootstrapped_samples):
    """Filtering original matrix by rows and columns to create the bootstrap matrix.
    Function is paralellized with numba and especially useful in case of big datasets, i.e., large distance matrices.

    :param M: Original matrix.
    :type M: pandas.DataFrame
    :param bootstrapped_samples: (sorted) bootstrapped samples for bootstrap matrix creation
    :type bootstrapped_samples: numpy array
    :return: M_bootstrapped: bootstrapped matrix
    :rtype: pandas.DataFrame
    """

    n = M.shape[0]

    M_bootstrapped = np.empty((n, n), dtype=M.dtype).T  # transpose to get F-contiguous

    for j in prange(n):
        M_bootstrapped[:, j] = M[:, bootstrapped_samples[j]][bootstrapped_samples]

    return M_bootstrapped


def _bootstrap_matrix(M):
    """Create a bootstrap from the original matrix.

    :param M: Original matrix.
    :type M: pandas.DataFrame
    :return: M_bootstrapped: bootstrapped matrix;
        mapping_bootstrapped_indices_to_original_indices: mapping from bootstrapped to original indices.
    :rtype: pandas.DataFrame, dict
    """

    lm = len(M)
    bootstrapped_samples = np.random.choice(np.arange(lm), lm)
    bootstrapped_samples = np.sort(
        bootstrapped_samples
    )  # Sort samples to increase speed. Does not affect downstream analysis because M is symmetric
    M_bootstrapped = _get_bootstrap(M, bootstrapped_samples)
    mapping_bootstrapped_indices_to_original_indices = {
        bootstrapped: original for bootstrapped, original in enumerate(bootstrapped_samples)
    }

    return M_bootstrapped, mapping_bootstrapped_indices_to_original_indices


def _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(labels, mapping=False):
    """Create dictionary that maps indices to cluster labels.

    :param labels: Clustering labels.
    :type labels: numpy.ndarray
    :param mapping: Mapping of bootstrapped to original indices, defaults to False
    :type mapping: bool, optional
    :return: Dictionary with cluster labels as keys and index of instances that belong to the respective cluster as labels.
    :rtype: dict
    """
    clusters = np.unique(labels)
    number_datapoints = len(labels)
    index_vector = np.arange(number_datapoints)

    indices_clusters = {}
    for cluster in clusters:
        indices = set(index_vector[labels == cluster])
        if mapping is not False:
            # translate from the bootstrapped indices to the original naming of the indices
            indices = set([mapping[index] for index in indices])

        indices_clusters[cluster] = indices

    return indices_clusters


def _compute_stability_indices(distance_matrix, cluster_method, clusters, indices_original_clusters):
    """Function that parallelizes the bootstrapping loop in the _compute_stability_indices function.
    Compute stability of each cluster via Jaccard Index of original clustering vs clustering of one bootstraped sample.

    :param distance_matrix: Proximity matrix of Random Forest model.
    :type distance_matrix: pandas.DataFrame
    :param cluster_method: Lambda function wrapping the k-mediods clustering function.
    :type cluster_method: object
    :param clusters: possible clusters (unique cluster labels)
    :type clusters: numpy array
    :param indices_original_clusters:  dictionary that maps indices to cluster labels
    :type indices_original_clusters: dict
    :return: Dictionary with Jaccard scores for each cluster in the given boostrap sample
    :rtype: dict
    """
    index_per_cluster = {cluster: 0 for cluster in clusters}

    (
        bootstrapped_distance_matrix,
        mapping_bootstrapped_indices_to_original_indices,
    ) = _bootstrap_matrix(distance_matrix)
    bootstrapped_labels = cluster_method(bootstrapped_distance_matrix)

    # now compute the indices for the different clusters
    indices_bootstrap_clusters = _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(
        bootstrapped_labels,
        mapping=mapping_bootstrapped_indices_to_original_indices,
    )
    jaccard_matrix = _compute_jaccard_matrix(clusters, indices_bootstrap_clusters, indices_original_clusters)

    # compute optimal jaccard index for each cluster -> choose maximum possible jaccard index first
    for cluster_round in range(len(jaccard_matrix)):
        best_index = jaccard_matrix.max(axis=1).max()
        original_cluster_number = jaccard_matrix.max(axis=1).argmax()
        bootstrapped_cluster_number = jaccard_matrix[original_cluster_number].argmax()
        jaccard_matrix[original_cluster_number] = -np.inf
        jaccard_matrix[:, bootstrapped_cluster_number] = -np.inf

        original_cluster = clusters[original_cluster_number]
        index_per_cluster[original_cluster] += best_index

    return index_per_cluster


def _compute_stability_indices_parallel(distance_matrix, labels, cluster_method, bootstraps, n_jobs):
    """Compute stability of each cluster via Jaccard Index of bootstraped vs original clustering.

    :param distance_matrix: Proximity matrix of Random Forest model.
    :type distance_matrix: pandas.DataFrame
    :param labels: original cluster labels
    :type labels: numpy array
    :param cluster_method: Lambda function wrapping the k-mediods clustering function.
    :type cluster_method: object
    :param bootstraps: Number of bootstraps to compute the Jaccard Index, defaults to 300
    :type bootstraps: int
    :param n_jobs: number of jobs to run in parallel when computing the cluster stability. n_jobs=1 means no parallel computing is used, defaults to 1
    :type n_jobs: int, optional
    :return: Dictionary with cluster labels as keys and Jaccard Indices as values.
    :rtype: dict
    """
    clusters = np.unique(labels)
    indices_original_clusters = _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(labels)

    # Compute Jaccard Index per bootstrapped sample
    index_per_cluster = Parallel(n_jobs=n_jobs)(
        delayed(_compute_stability_indices)(
            distance_matrix, cluster_method, clusters, indices_original_clusters
        )
        for i in range(bootstraps)
    )
    # Sum Jaccard values of the same keys across dictionaries
    index_per_cluster = dict(functools.reduce(operator.add, map(collections.Counter, index_per_cluster)))
    # normalize:
    index_per_cluster = {cluster: index_per_cluster[cluster] / bootstraps for cluster in clusters}

    return index_per_cluster


def optimizeK(
    distance_matrix,
    y,
    model_type,
    max_K,
    method_clustering,
    init_clustering,
    max_iter_clustering,
    discart_value_JI,
    bootstraps_JI,
    random_state,
    n_jobs,
    verbose,
):
    """Compute the optimal number of clusters for k-medoids clustering (trade-off between cluster purity and cluster stability).

    :param distance_matrix: Proximity matrix of Random Forest model.
    :type distance_matrix: pandas.DataFrame
    :param y: Target column.
    :type y: pandas.Series
    :param model_type: Model type of Random Forest model: classifier or regression.
    :type model_type: str
    :param max_K: Maximum number of clusters for cluster score computation, defaults to 6
    :type max_K: int
    :param method_clustering: Which algorithm to use. 'alternate' is faster while 'pam' is more accurate, defaults to 'pam'
    :type method_clustering: {'alternate', 'pam'}, optional
    :param init_clustering: Specify medoid initialization method. To speed up computation for large datasets use 'random'.
        See sklearn documentation for parameter description, defaults to 'k-medoids++'
    :type init_clustering: {'random', 'heuristic', 'k-medoids++', 'build'}, optional
    :param max_iter_clustering: Number of iterations for k-medoids clustering, defaults to 500
    :type max_iter_clustering: int
    :param discart_value: Minimum Jaccard Index for cluster stability, defaults to 0.6
    :type discart_value: float
    :param bootstraps_JI: Number of bootstraps to compute the Jaccard Index, defaults to 300
    :type bootstraps_JI: int
    :param random_state: Seed number for random state, defaults to 42
    :type random_state: int
    :param n_jobs: number of jobs to run in parallel when computing the cluster stability.
        n_jobs=1 means no parallel computing is used, defaults to 1
    :type n_jobs: int, optional
    :return: Optimal number of clusters.
    :rtype: int
    :param verbose: print the output of fgc cluster optimization process (the Jaccard index and score for each cluster number); defaults to 1 (printing). Set to 0 for no outputs.
    :type verbose: {0,1}, optional
    """
    np.random.seed(random_state)

    # Check distance matrix
    matrix_shape = distance_matrix.shape
    assert len(matrix_shape) == 2, "error distance_matrix is not a matrix"
    assert matrix_shape[0] == matrix_shape[1], "error distance matrix is not square"

    score_min = np.inf
    optimal_k = 1
    disable = True if verbose == 0 else False

    for k in tqdm(range(2, max_K), disable=disable):
        # compute clusters
        cluster_method = (
            lambda X: kmedoids.KMedoids(
                n_clusters=k,
                method=method_clustering,
                init=init_clustering,
                metric="precomputed",
                max_iter=max_iter_clustering,
                random_state=random_state,
            )
            .fit(X)
            .labels_
        )
        labels = cluster_method(distance_matrix)

        # compute jaccard indices
        index_per_cluster = _compute_stability_indices_parallel(
            distance_matrix, labels, cluster_method, bootstraps_JI, n_jobs
        )
        min_index = min([index_per_cluster[cluster] for cluster in index_per_cluster.keys()])

        # only continue if jaccard indices are all larger than discart_value_JI (thus all clusters are stable)
        if not disable:
            print("For number of cluster {} the Jaccard Index is {}".format(k, min_index))
        if min_index > discart_value_JI:
            if model_type == "classification":
                # compute balanced purities
                score = statistics.compute_balanced_average_impurity(y, labels)
            elif model_type == "regression":
                # compute the total within cluster variation
                score = statistics.compute_total_within_cluster_variation(y, labels)
            if score < score_min:
                optimal_k = k
                score_min = score

            if not disable:
                print("For number of cluster {} the score is {}".format(k, score))
        else:
            if not disable:
                print("Clustering is instable, no score computed!")

    return optimal_k



############################################
# imports
############################################

import pandas as pd

from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier

from fgclustering.utils import *


############################################
# Tests
############################################


def test_proximityMatrix():
    # test data
    X, y = make_classification(
        n_samples=300,
        n_features=10,
        n_informative=4,
        n_redundant=2,
        n_classes=2,
        n_clusters_per_class=1,
        random_state=1,
    )
    X = pd.DataFrame(X)

    model = RandomForestClassifier(
        max_depth=10,
        max_features="sqrt",
        max_samples=0.8,
        bootstrap=True,
        oob_score=True,
        random_state=42,
    )
    model.fit(X, y)

    result = proximityMatrix(model, X)

    dim1, dim2 = result.shape

    assert dim1 == dim2, "error: proximity matrix not quadratic"
    assert dim1 == len(X), "error: proximity matrix has wrong dimensions"

    assert (
        np.diag(result).min() == 1.0
    ), "error: proximity matrix should have ones on diagonal"
    assert (
        np.diag(result).max() == 1.0
    ), "error: proximity matrix should have ones on diagonal"

    assert np.allclose(result, result.T), "error: proximity matrix should be symmetric"



############################################
# imports
############################################

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from sklearn.preprocessing import MinMaxScaler, LabelEncoder

import fgclustering.utils as utils


############################################
# functions
############################################


def _plot_feature_importance(
    p_value_of_features_ranked: pd.DataFrame,
    p_value_of_features_per_cluster: pd.DataFrame,
    thr_pvalue: float,
    top_n: int,
    num_cols: int,
    save: str,
):
    """
    Generate and display a plot showing the importance of features based on p-values.
    The plot includes both global feature importance and local feature importance for each cluster.
    Global importance is based on all clusters combined, while local importance is specific to each cluster.

    :param p_value_of_features_ranked: DataFrame containing p-values of features, ranked by p-value.
    :type p_value_of_features_ranked: pandas.DataFrame
    :param p_value_of_features_per_cluster: DataFrame containing p-values of features for each cluster.
    :type p_value_of_features_per_cluster: pandas.DataFrame
    :param thr_pvalue: P-value threshold for display. Only features with p-values below this threshold
                    are considered significant. Defaults to 1 (no filtering).
    :type thr_pvalue: float, optional
    :param top_n: Number of top features to display in the plot. If None, all features are included.
                Defaults to None.
    :type top_n: int, optional
    :param num_cols: Number of plots per row in the output figure. Defaults to 4.
    :type num_cols: int, optional
    :param save: Filename to save the plot. If None, the plot will not be saved. Defaults to None.
    :type save: str, optional
    """

    # Determine figure size dynamically based on the number of features
    num_features = len(p_value_of_features_ranked.columns)
    figsize_width = 6.5
    figsize_height = max(figsize_width, int(np.ceil(5 * num_features / 25)))

    num_subplots = 1 + p_value_of_features_per_cluster.shape[1]
    num_cols = min(num_cols, num_subplots)
    num_rows = int(np.ceil(num_subplots / num_cols))

    plt.figure(figsize=(num_cols * figsize_width, num_rows * figsize_height))
    plt.subplots_adjust(top=0.95, hspace=0.8, wspace=0.8)
    plt.suptitle(
        f"Feature Importance - Showing {'top ' + str(top_n) if top_n else 'all'} features",
        fontsize=14,
    )
    sns.set_theme(style="whitegrid")

    # Plot global feature importance
    importance_global = pd.DataFrame(
        {
            "Feature": p_value_of_features_ranked.columns,
            "Importance": utils.log_transform(p_value_of_features_ranked.loc["p_value"].to_list()),
        }
    ).sort_values(by="Importance", ascending=False)

    ax = plt.subplot(num_rows, num_cols, 1)
    sns.barplot(data=importance_global, x="Importance", y="Feature", color="#3470a3")
    ax.axvline(
        x=utils.log_transform(thr_pvalue), color="red", linestyle="--", label=f"thr p-value = {thr_pvalue}"
    )
    ax.set_title(f"Cluster all")
    ax.legend(bbox_to_anchor=(1, 1), loc=2)

    # Plot local feature importance
    for n, cluster in enumerate(p_value_of_features_per_cluster.columns):
        importance_local = pd.DataFrame(
            {
                "Feature": p_value_of_features_per_cluster.index,
                "Importance": utils.log_transform(p_value_of_features_per_cluster[cluster].to_list()),
            }
        ).sort_values(by="Importance", ascending=False)
        ax = plt.subplot(num_rows, num_cols, n + 2)
        sns.barplot(data=importance_local, x="Importance", y="Feature", color="#3470a3")
        ax.axvline(x=utils.log_transform(thr_pvalue), color="red", linestyle="--")
        ax.set_title(f"Cluster {cluster}")
        # ax.legend(bbox_to_anchor=(1, 1), loc=2)

    plt.tight_layout(rect=[0, 0, 1, 0.95])

    if save:
        plt.savefig(f"{save}_feature_importance.png", bbox_inches="tight", dpi=300)
    plt.show()


def _plot_distributions(
    data_clustering_ranked: pd.DataFrame, thr_pvalue: float, top_n: int, num_cols: int, save: str
):
    """Plot feature boxplots (for continuous features) or barplots (for categorical features) divided by clusters,
    where features are filtered and ranked by p-value of a statistical test (ANOVA for continuous features,
    chi square for categorical features).

    :param data_clustering_ranked: Filtered and ranked data frame incl features, target and cluster numbers.
    :type data_clustering_ranked: pandas.DataFrame
    :param thr_pvalue: P-value threshold used for feature filtering
    :type thr_pvalue: float, optional
    :param num_cols: Number of plots in one row.
    :type num_cols: int
    :param save: Filename to save plot.
    :type save: str
    """
    features_to_plot = data_clustering_ranked.drop("cluster", axis=1, inplace=False).columns.to_list()

    num_rows = int(np.ceil(len(features_to_plot) / num_cols))

    plt.figure(figsize=(num_cols * 4.5, num_rows * 4.5))
    plt.subplots_adjust(top=0.95, hspace=0.8, wspace=0.8)
    plt.suptitle(
        f"Distribution of feature values across subgroups - Showing {'top ' + str(top_n) if top_n else 'all'} features with p-value < {thr_pvalue}",
        fontsize=14,
    )

    for n, feature in enumerate(features_to_plot):
        # add a new subplot iteratively
        ax = plt.subplot(num_rows, num_cols, n + 1)
        if data_clustering_ranked[feature].nunique() < 5 or isinstance(
            data_clustering_ranked[feature].dtype, pd.CategoricalDtype
        ):
            sns.countplot(
                x="cluster",
                hue=feature,
                data=data_clustering_ranked,
                ax=ax,
                palette="Blues_r",
            )
            ax.set_title(f"Feature: {feature}")
            ax.legend(bbox_to_anchor=(1, 1), loc=2)
        else:
            sns.boxplot(
                x="cluster",
                y=feature,
                data=data_clustering_ranked,
                ax=ax,
                color="#3470a3",
            )
            ax.set_title(f"Feature: {feature}")

    plt.tight_layout(rect=[0, 0, 1, 0.95])

    if save:
        plt.savefig(f"{save}_boxplots.png", bbox_inches="tight", dpi=300)
    plt.show()


def _plot_heatmap_classification(
    data_clustering_ranked: pd.DataFrame,
    thr_pvalue: float,
    top_n: int,
    heatmap_type: str,
    save: str,
):
    """
    Generates a classification heatmap visualization for clustered data, supporting both static and interactive plots.
    Displays target and feature heatmaps with cluster boundaries.

    :param data_clustering_ranked: A DataFrame containing the data for clustering, including target and cluster labels.
    :type data_clustering_ranked: pd.DataFrame
    :param thr_pvalue: Threshold for feature p-value significance in the heatmap.
    :type thr_pvalue: float
    :param top_n: Number of top features to display in the heatmap, or None to display all features.
    :type top_n: int
    :param heatmap_type: Type of heatmap to generate: "static" for Matplotlib or "interactive" for Plotly.
    :type heatmap_type: str
    :param save: File path for saving the heatmap. Only supported for static heatmaps.
    :type save: str
    :raises RuntimeError: If `heatmap_type` is "interactive" and `save` is specified, as saving interactive plots is not implemented.
    :return: None
    """
    cluster_labels = data_clustering_ranked["cluster"]

    # Encode categorical target for plotting and transpose for plotting
    target_encoder = LabelEncoder()
    target = target_encoder.fit_transform(data_clustering_ranked["target"])
    target = pd.DataFrame([target], index=["target"])

    categories = target_encoder.classes_

    # Process features and transpose for plotting
    features = _process_features_for_heatmap(data_clustering_ranked.drop(columns=["target", "cluster"]))
    features = features.T

    # Determine cluster boundaries for separator lines
    boundaries = np.where(np.diff(cluster_labels) != 0)[0] + 1

    target_color, features_color, boundaries_color, boundaries_width, title = _get_heatmap_plotting_settings(
        target, top_n, thr_pvalue
    )

    if heatmap_type == "static":
        # Get plotting settings

        color_palette = sns.color_palette(target_color, n_colors=len(categories))
        cmap_target = sns.color_palette(color_palette, as_cmap=True)
        color_map = {i: color_palette[i] for i in range(len(categories))}

        fig, ax_target, ax_target_cb, ax_features, ax_features_cb, target_plot, feature_plot = (
            _plot_heatmaps_static(
                target,
                cmap_target,
                features,
                features_color,
                boundaries,
                boundaries_color,
                boundaries_width,
                title,
            )
        )

        # Add a custom legend or color bar for targets plot
        handles = [
            plt.Line2D([0], [0], marker="o", color="w", markerfacecolor=color_map[i], markersize=10)
            for i in range(len(categories))
        ]
        ax_features.legend(
            handles,
            categories,
            title="Target Categories",
            loc="lower center",
            bbox_to_anchor=(0.5, -0.3),
            ncol=len(categories),
        )

        plt.tight_layout()
        if save:
            plt.savefig(f"{save}_heatmap.png", bbox_inches="tight", dpi=300)
        plt.show()

    elif heatmap_type == "interactive":
        # Create a color scale for the target categories
        color_palette_rgb = [
            f"rgb({int(r * 255)},{int(g * 255)},{int(b * 255)})"
            for r, g, b in sns.color_palette(target_color, n_colors=len(categories))
        ]
        color_map = {i: color_palette_rgb[i] for i in range(len(categories))}
        colorscale = [[i / (len(categories) - 1), color_map[i]] for i in range(len(categories))]

        fig = _plot_heatmaps_interactive(
            target,
            colorscale,
            None,
            False,
            features,
            features_color,
            boundaries,
            boundaries_color,
            boundaries_width,
            title,
        )
        for i, category in enumerate(categories):
            fig.add_trace(
                go.Scatter(
                    x=[None],
                    y=[None],
                    mode="markers",
                    marker=dict(size=10, color=color_map[i]),
                    legendgroup=category,
                    showlegend=True,
                    name=category,
                )
            )
        fig.show()

        if save:
            raise RuntimeError(
                "Saving interactive plots is not implemented. Please set heatmap_type='static' to save the plot."
            )


def _plot_heatmap_regression(
    data_clustering_ranked: pd.DataFrame,
    thr_pvalue: float,
    top_n: int,
    heatmap_type: str,
    save: str,
):
    """
    Generates a regression heatmap visualization for clustered data, with options for static or interactive plots.
    Displays target and feature heatmaps with cluster boundaries.

    :param data_clustering_ranked: A DataFrame containing the data for clustering, including target and cluster labels.
    :type data_clustering_ranked: pd.DataFrame
    :param thr_pvalue: Threshold for feature p-value significance in the heatmap.
    :type thr_pvalue: float
    :param top_n: Number of top features to display in the heatmap, or None to display all features.
    :type top_n: int
    :param heatmap_type: Type of heatmap to generate: "static" for Matplotlib or "interactive" for Plotly.
    :type heatmap_type: str
    :param save: File path for saving the heatmap. Only supported for static heatmaps.
    :type save: str
    :raises RuntimeError: If `heatmap_type` is "interactive" and `save` is specified, as saving interactive plots is not implemented.
    :return: None
    """
    cluster_labels = data_clustering_ranked["cluster"]

    # Get traget and transpose for plotting
    target = data_clustering_ranked["target"]
    target = pd.DataFrame([target], index=["target"])

    # Process features and transpose for plotting
    features = _process_features_for_heatmap(data_clustering_ranked.drop(columns=["target", "cluster"]))
    features = features.T

    # Determine cluster boundaries for separator lines
    boundaries = np.where(np.diff(cluster_labels) != 0)[0] + 1
    target_color, features_color, boundaries_color, boundaries_width, title = _get_heatmap_plotting_settings(
        target, top_n, thr_pvalue
    )

    if heatmap_type == "static":
        # Get plotting settings
        cmap_target = sns.color_palette(target_color, as_cmap=True)

        # Plot heatmaps
        fig, ax_target, ax_target_cb, ax_features, ax_features_cb, target_plot, feature_plot = (
            _plot_heatmaps_static(
                target,
                cmap_target,
                features,
                features_color,
                boundaries,
                boundaries_color,
                boundaries_width,
                title,
            )
        )

        # Add a custom legend or color bar for targets plot
        cbar = fig.colorbar(target_plot.collections[0], ax=ax_target_cb, orientation="vertical", pad=0.1)
        cbar.set_label("Target")

        plt.tight_layout()
        if save:
            plt.savefig(f"{save}_heatmap.png", bbox_inches="tight", dpi=300)
        plt.show()

    elif heatmap_type == "interactive":
        target_colorbar = dict(title="Target Scale", x=1.2)

        fig = _plot_heatmaps_interactive(
            target,
            target_color,
            target_colorbar,
            True,
            features,
            features_color,
            boundaries,
            boundaries_color,
            boundaries_width,
            title,
        )
        fig.show()

        if save:
            raise RuntimeError(
                "Saving interactive plots is not implemented. Please set heatmap_type='static' to save the plot."
            )


def _process_features_for_heatmap(features):
    """
    Processes a DataFrame of features to prepare them for heatmap visualization.
    Categorical features are encoded as numeric, and all features are normalized to a 0-1 range.

    :param features: A DataFrame containing the features to process.
    :type features: pd.DataFrame
    :return: A DataFrame with categorical features encoded as numeric and all features normalized.
    :rtype: pd.DataFrame
    """
    # Encode categorical features as numeric
    for col in features.columns:
        if pd.api.types.is_string_dtype(features[col]):
            features[col] = features[col].astype("category")
            features[col] = features[col].cat.codes

    # Normalize all features (both encoded categorical and continuous)
    scaler = MinMaxScaler()
    scaler.set_output(transform="pandas")
    features = scaler.fit_transform(features)

    return features


def _get_heatmap_plotting_settings(target, top_n, thr_pvalue):
    """
    Configures plotting settings for heatmap visualizations, including color schemes,
    boundary thickness, and plot title.

    :param target: Target data, i.e. a DataFrame containing the target values.
    :type target: pd.DataFrame
    :param top_n: Number of top features to display, or None to display all features.
    :type top_n: int or None
    :param thr_pvalue: Threshold for p-value significance in feature selection.
    :type thr_pvalue: float
    :return: Tuple containing color settings for target and features, boundary color, boundary length, and plot title.
    :rtype: tuple(str, str, str, int, str)
    """
    color_target = "Greens"
    color_features = "coolwarm"
    boundaries_color = "white"

    boundaries_width = int(np.ceil(np.log(target.shape[1])))

    title = f"Subgroups of instances that follow similar decision paths in the RF model \n Showing {'top ' + str(top_n) if top_n else 'all'} features with p-value < {thr_pvalue}"

    return color_target, color_features, boundaries_color, boundaries_width, title


def _plot_heatmaps_static(
    target, cmap_target, features, features_color, boundaries, boundaries_color, boundaries_width, title
):
    """
    Creates static heatmaps for target and features with cluster boundaries.

    :param target: DataFrame representing the target values to plot.
    :type target: pd.DataFrame
    :param cmap_target: Colormap for the target heatmap.
    :type cmap_target: str or matplotlib.colors.Colormap
    :param features: DataFrame representing the feature values to plot.
    :type features: pd.DataFrame
    :param features_color: Colormap for the feature heatmap.
    :type features_color: str or matplotlib.colors.Colormap
    :param boundaries: List of boundary positions for separating clusters in the heatmaps.
    :type boundaries: list of int
    :param boundaries_color: Color of the boundary lines.
    :type boundaries_color: str
    :param boundaries_width: Width of the boundary lines.
    :type boundaries_width: float
    :param title: Title for the heatmap plot.
    :type title: str
    :return: Tuple containing the figure, target axis, target color bar axis, features axis, features color bar axis,
             target heatmap plot, and features heatmap plot.
    :rtype: tuple(matplotlib.figure.Figure, matplotlib.axes.Axes, matplotlib.axes.Axes, matplotlib.axes.Axes,
                  matplotlib.axes.Axes, sns.heatmap, sns.heatmap)
    """
    # Set up the figure and subplots
    figure_size = max(6.5, int(np.ceil(5 * len(features) / 25)))
    fig, axes = plt.subplots(
        nrows=2,
        ncols=2,
        sharex=True,
        height_ratios=[1, 9],
        width_ratios=[20, 1],
        figsize=(2.5 * figure_size, figure_size),
    )

    # Add extra axis for legends and disable plot in this axis
    ax_target, ax_target_cb = axes[0]
    ax_features, ax_features_cb = axes[1]

    ax_target_cb.axis("off")
    ax_features_cb.axis("off")

    # Adjust the axes
    ax_target.set_title(title)
    ax_target.set_xticks([])

    ax_features.set_xlabel("Samples")
    ax_features.set_xticks([])

    # Plot the target heatmap
    target_plot = sns.heatmap(
        target,
        ax=ax_target,
        cmap=cmap_target,
        cbar=False,
        yticklabels=target.index,
        xticklabels=False,
    )
    ax_target.set_yticklabels(ax_target.get_yticklabels(), rotation=0)

    # Plot the feature heatmap
    feature_plot = sns.heatmap(
        features,
        ax=ax_features,
        cmap=features_color,
        cbar=False,
        yticklabels=features.index,
        xticklabels=False,
    )
    # Add color bar feature plot
    cbar = fig.colorbar(feature_plot.collections[0], ax=ax_features_cb, orientation="vertical", pad=0.1)
    cbar.set_label("Features")

    # Add cluster separators
    for ax in [ax_target, ax_features]:
        for boundary in boundaries:
            ax.axvline(boundary, color=boundaries_color, lw=boundaries_width)

    return fig, ax_target, ax_target_cb, ax_features, ax_features_cb, target_plot, feature_plot


def _plot_heatmaps_interactive(
    target,
    target_colorscale,
    target_colorbar,
    target_showscale,
    features,
    features_color,
    boundaries,
    boundaries_color,
    boundaries_width,
    title,
):
    """
    Create an interactive Plotly heatmap visualization for target and features.

    :param target: Target data as a DataFrame.
    :param target_colorscale: Colorscale for the target heatmap.
    :param target_colorbar: Configuration for the target heatmap colorbar.
    :param target_showscale: Whether to display the color scale for the target heatmap.
    :param features: Features data as a DataFrame.
    :param features_color: Colorscale for the features heatmap.
    :param boundaries: List of boundary positions for separators.
    :param boundaries_color: Color of the boundary lines.
    :param boundaries_width: Width of the boundary lines.
    :param title: Title of the heatmap plot.
    :return: A Plotly figure object.
    """
    fig = make_subplots(rows=2, cols=1, row_heights=[0.1, 0.9], shared_xaxes=True, vertical_spacing=0.02)

    # Target heatmap (top row) with categorical colors
    fig.add_trace(
        go.Heatmap(
            z=target.values,  # Use only the target values
            y=target.index,
            colorscale=target_colorscale,
            colorbar=target_colorbar,
            showscale=target_showscale,  # Hide color bar for the target
        ),
        row=1,
        col=1,
    )

    # Reverse order the rows of the dataframe to plot in correct order
    features = features.iloc[::-1]

    # Continuous feature heatmap
    fig.add_trace(
        go.Heatmap(
            z=features.values,
            y=features.index,
            colorscale=utils.matplotlib_to_plotly(features_color),
            colorbar=dict(title="Features", x=1.1),  # Adjust position of color bar
            showscale=True,
        ),
        row=2,
        col=1,
    )

    # Add separators for clusters
    for xref, yref, y1 in [("x1", "y1", 0.5), ("x2", "y2", len(features.index) - 0.5)]:
        for boundary in boundaries:
            fig.add_shape(
                type="line",
                x0=boundary - 0.5,
                x1=boundary - 0.5,
                y0=-0.5,
                y1=y1,
                line=dict(color=boundaries_color, width=boundaries_width),
                xref=xref,
                yref=yref,
            )

    # Add axis title and layout adjustments
    fig.add_annotation(
        x=0.5,
        y=-0.2,
        showarrow=False,
        text="Samples",
        xref="paper",
        yref="paper",
        font=dict(size=16),
    )
    fig.update_layout(
        title=dict(text=title, y=0.95, x=0.5, xanchor="center", yanchor="top"),
        legend=dict(yanchor="top", y=0.9, xanchor="right", x=1.3),
    )

    return fig



############################################
# imports
############################################

import matplotlib.colors
import numpy as np
from numba import njit, prange

import matplotlib

############################################
# functions
############################################


def log_transform(p_values: list, epsilon: float = 1e-50):
    """
    Apply a log transformation to p-values to enhance numerical stability and highlight differences.
    Adds a small constant `epsilon` to avoid taking the log of zero and normalizes by dividing by
    the log of `epsilon`.

    :param p_values: List of p-values to be transformed.
    :type p_values: list
    :param epsilon: Small constant added to p-values to avoid log of zero. Defaults to 1e-50.
    :type epsilon: float, optional
    :return: Transformed p-values after log transformation.
    :rtype: numpy.ndarray
    """
    # add a small constant epsilon
    p_values = np.clip(p_values, epsilon, 1)
    return -np.log(p_values) / -np.log(epsilon)


@njit
def _calculate_proximityMatrix(terminals, normalize):
    """Calculate proximity matrix given leaf indices from the random forest model.
    Function is paralellized with numba and especially useful in case of big datasets or forests with a large number of estimators

    :param terminals: ndarray of shape (n_samples, n_estimators), result of apply() method of RandomForest; for each tree in the forest, it contais leaf indices a sample ended up in.
    :type terminals: numpy array
    :param normalize: Normalize proximity matrix by number of trees in the Random Forest, defaults to True.
    :type normalize: bool, optional
    :return: calculated proximity matrix of Random Forest model
    :rtype: numpy array
    """

    n = terminals.shape[0]
    proxMat = np.zeros((n, n))
    for i in prange(n):
        for j in prange(i, n):
            proxMat[i, j] = np.sum(terminals[i, :] == terminals[j, :])
    proxMat = proxMat + proxMat.T - np.eye(n) * proxMat[0, 0]

    if normalize:
        proxMat = proxMat / terminals.shape[1]

    return proxMat


def proximityMatrix(model, X, normalize=True):
    """Calculate proximity matrix of Random Forest model.

    :param model: Trained Random Forest model.
    :type model: sklearn.ensemble
    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param normalize: Normalize proximity matrix by number of trees in the Random Forest, defaults to True.
    :type normalize: bool, optional
    :return: Proximity matrix of Random Forest model.
    :rtype: numpy array
    """

    terminals = model.apply(X)

    return _calculate_proximityMatrix(terminals, normalize)


def matplotlib_to_plotly(cmap_name: str, pl_entries: int = 255):
    """
    Converts a Matplotlib colormap to a Plotly colorscale.

    :param cmap_name: Name of the Matplotlib colormap.
    :type cmap_name: str
    :param pl_entries: Number of color entries in the Plotly colorscale.
    :type pl_entries: int
    :return: A Plotly-compatible colorscale
    :rtype: list
    """
    cmap = matplotlib.colormaps.get_cmap(cmap_name)
    h = np.linspace(0, 1, pl_entries)
    colors = cmap(h)[:, :3]
    colors = [matplotlib.colors.rgb2hex(color) for color in colors]
    colorscale = [[i / (pl_entries - 1), color] for i, color in enumerate(colors)]
    return colorscale



############################################
# imports
############################################

import numpy as np
import pandas as pd
import kmedoids
import fgclustering.utils as utils
import fgclustering.optimizer as optimizer
import fgclustering.plotting as plotting
import fgclustering.statistics as stats
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from typing import Union

import warnings

############################################
# Forest-guided Clustering
############################################


class FgClustering:
    """
    Forest-Guided Clustering.

    Computes feature importance based on subgroups of instances that follow similar decision
    rules within a Random Forest model. This class is designed to handle both regression and
    classification problems using a trained Random Forest model.

    :param model: Trained Random Forest model, which must be an instance of
                  `sklearn.ensemble.RandomForestRegressor` or
                  `sklearn.ensemble.RandomForestClassifier`.
    :type model: sklearn.ensemble.RandomForestRegressor or sklearn.ensemble.RandomForestClassifier
    :param data: Input data containing the feature matrix. If `target_column` is a string,
                 it should be a column name in the `data` DataFrame.
    :type data: pandas.DataFrame
    :param target_column: Name of the target column as a string, or target values as a numpy array or pd.Series.
                          If provided as a string, it must correspond to a column in the `data` DataFrame.
    :type target_column: str or numpy.ndarray or pd.Series
    :param random_state: Seed for the random number generator, used for reproducibility. Defaults to 42.
    :type random_state: int, optional
    :raises ValueError: Raised if `model` is not an instance of `sklearn.ensemble.RandomForestRegressor`
                        or `sklearn.ensemble.RandomForestClassifier`.
    """

    def __init__(
        self,
        model: Union[RandomForestClassifier, RandomForestRegressor],
        data: pd.DataFrame,
        target_column: Union[str, np.ndarray, pd.Series],
        random_state: int = 42,
    ):
        self.random_state = random_state

        # check if random forest is regressor or classifier
        if "RandomForestRegressor" in str(type(model)):
            self.model_type = "regression"
            print("Interpreting RandomForestRegressor")
        elif "RandomForestClassifier" in str(type(model)):
            self.model_type = "classification"
            print("Interpreting RandomForestClassifier")
        else:
            raise ValueError(
                f"Do not recognize {str(type(model))}. Can only work with sklearn RandomForestRegressor or RandomForestClassifier."
            )

        if type(target_column) == str:
            self.y = data.loc[:, target_column]
            self.X = data.drop(columns=[target_column])
        else:
            self.y = pd.Series(target_column)
            self.X = data

        self.y.reset_index(inplace=True, drop=True)
        self.X.reset_index(inplace=True, drop=True)

        self.proximity_matrix = utils.proximityMatrix(model, self.X)
        self.distance_matrix = 1 - self.proximity_matrix
        self.k = None
        self.cluster_labels = None

    def run(
        self,
        k: int = None,
        max_K: int = 8,
        method_clustering: str = "pam",
        init_clustering: str = "random",
        max_iter_clustering: int = 100,
        discart_value_JI: float = 0.6,
        bootstraps_JI: int = 100,
        bootstraps_p_value: int = 100,
        n_jobs: int = 1,
        verbose: int = 1,
    ):
        """Runs the forest-guided clustering model. The optimal number of clusters for a k-medoids clustering is computed,
        based on the distance matrix computed from the Random Forest proximity matrix.

        :param number_of_clusters: Number of clusters for the k-medoids clustering.
            Leave None if number of clusters should be optimized, defaults to None
        :type number_of_clusters: int, optional
        :param max_K: Maximum number of clusters for cluster score computation, defaults to 8
        :type max_K: int, optional
        :param method_clustering: Which algorithm to use. 'alternate' is faster while 'pam' is more accurate, defaults to 'pam'. Use 'fasterpam' for big datasets. See python kmedoids documentation for other implemented methods.
        :type method_clustering: {'fasterpam', 'fastpam1', 'pam', 'alternate', 'fastermsc', 'fastmsc', 'pamsil', and 'pammedsil'}, optional
        :param init_clustering: Specify medoid initialization method. See python kmedoids documentation for parameter description, defaults to 'random'
        :type init_clustering: {'random', 'first', 'build'}, optional
        :param max_iter_clustering: Number of iterations for k-medoids clustering, defaults to 100
        :type max_iter_clustering: int, optional
        :param discart_value_JI: Minimum Jaccard Index for cluster stability, defaults to 0.6
        :type discart_value_JI: float, optional
        :param bootstraps_JI: Number of bootstraps to compute the Jaccard Index, defaults to 100
        :type bootstraps_JI: int, optional
        :param bootstraps_p_value: Number of bootstraps to compute the p-value of feature importance, defaults to 100
        :type bootstraps_p_value: int, optional
        :param n_jobs: maximum number of jobs to run in parallel when creating bootstraps to compute the Jaccard index.
            n_jobs=1 means no parallel computing is used, defaults to 1
        :type n_jobs: int, optional
        :param verbose: print the output of fgc cluster optimization process (the Jaccard index and score for each cluster number); defaults to 1 (printing). Set to 0 for no outputs.
        :type verbose: {0,1}, optional
        """
        """
        Runs the forest-guided clustering model to compute the optimal number of clusters using k-medoids clustering,
        based on the distance matrix derived from the Random Forest proximity matrix. The method can either optimize 
        the number of clusters or use a specified number.

        :param number_of_clusters: Number of clusters for the k-medoids clustering. If `None`, the number of clusters 
                                    will be optimized based on the distance matrix, defaults to None.
        :type number_of_clusters: int, optional
        :param max_K: Maximum number of clusters to consider when computing cluster scores. Used if `number_of_clusters` 
                    is not provided. Defaults to 8.
        :type max_K: int, optional
        :param method_clustering: Clustering algorithm to use. Options include 'fasterpam', 'fastpam1', 'pam', 'alternate', 
                                'fastermsc', 'fastmsc', 'pamsil', and 'pammedsil'. Defaults to 'pam'. Use 'fasterpam' 
                                for larger datasets. Refer to the k-medoids documentation for more details.
        :type method_clustering: {'fasterpam', 'fastpam1', 'pam', 'alternate', 'fastermsc', 'fastmsc', 'pamsil', 
                                'pammedsil'}, optional
        :param init_clustering: Method for initializing medoids. Options include 'random', 'first', and 'build'. Defaults 
                                to 'random'. See the k-medoids documentation for more information.
        :type init_clustering: {'random', 'first', 'build'}, optional
        :param max_iter_clustering: Maximum number of iterations for the k-medoids algorithm, defaults to 100.
        :type max_iter_clustering: int, optional
        :param discart_value_JI: Minimum Jaccard Index value for determining cluster stability. Clusters with Jaccard 
                                Index below this value are discarded, defaults to 0.6.
        :type discart_value_JI: float, optional
        :param bootstraps_JI: Number of bootstrap iterations to compute the Jaccard Index, defaults to 100.
        :type bootstraps_JI: int, optional
        :param bootstraps_p_value: Number of bootstrap iterations to compute the p-value for feature importance, defaults 
                                    to 100.
        :type bootstraps_p_value: int, optional
        :param n_jobs: Number of parallel jobs to run when computing the Jaccard Index bootstraps. Defaults to 1, meaning 
                    no parallel computation.
        :type n_jobs: int, optional
        :param verbose: Verbosity level for output. If set to 1, prints the optimization process including Jaccard Index 
                        and scores for each number of clusters. If set to 0, no output is printed. Defaults to 1.
        :type verbose: {0, 1}, optional
        """
        if k is None:
            self.k = optimizer.optimizeK(
                distance_matrix=self.distance_matrix,
                y=self.y.to_numpy(),
                model_type=self.model_type,
                max_K=max_K,
                method_clustering=method_clustering,
                init_clustering=init_clustering,
                max_iter_clustering=max_iter_clustering,
                discart_value_JI=discart_value_JI,
                bootstraps_JI=bootstraps_JI,
                random_state=self.random_state,
                n_jobs=n_jobs,
                verbose=verbose,
            )

            if self.k == 1:
                warnings.warn("No stable clusters were found!")
                return

            print(f"Optimal number of cluster is: {self.k}")

        else:
            self.k = k
            print(f"Use {self.k} as number of cluster")

        self.cluster_labels = (
            kmedoids.KMedoids(
                n_clusters=self.k,
                method=method_clustering,
                init=init_clustering,
                metric="precomputed",
                max_iter=max_iter_clustering,
                random_state=self.random_state,
            )
            .fit(self.distance_matrix)
            .labels_
        )

        self.data_clustering_ranked, self.p_value_of_features_ranked = (
            stats.calculate_global_feature_importance(
                X=self.X, y=self.y, cluster_labels=self.cluster_labels, model_type=self.model_type
            )
        )

        self.p_value_of_features_per_cluster = stats.calculate_local_feature_importance(
            data_clustering_ranked=self.data_clustering_ranked, bootstraps_p_value=bootstraps_p_value
        )

    def calculate_statistics(self, data, target_column, bootstraps_p_value=100):
        """
        Recalculates p-values for each feature based on the new feature matrix, affecting all related plotting functions.
        The new feature matrix must have the same number of samples and the same ordering of samples as the original matrix.

        :param data: Input data containing the new feature matrix. If `target_column` is a string,
                     it should be a column name in the `data` DataFrame.
        :type data: pandas.DataFrame
        :param target_column: Name of the target column as a string, or target values as a numpy array or pd.Series.
                              If provided as a string, it must correspond to a column in the `data` DataFrame.
        :type target_column: str or numpy.ndarray or pd.Series
        :param bootstraps_p_value: Number of bootstraps to use for calculating the p-value of feature importance. Defaults
                                    to 100.
        :type bootstraps_p_value: int, optional
        """
        if type(target_column) == str:
            y = data.loc[:, target_column]
            X = data.drop(columns=[target_column])
        else:
            y = pd.Series(target_column)
            X = data

        y.reset_index(inplace=True, drop=True)
        X.reset_index(inplace=True, drop=True)

        self.data_clustering_ranked, self.p_value_of_features_ranked = (
            stats.calculate_global_feature_importance(
                X=X, y=y, cluster_labels=self.cluster_labels, model_type=self.model_type
            )
        )
        self.p_value_of_features_per_cluster = stats.calculate_local_feature_importance(
            data_clustering_ranked=self.data_clustering_ranked, bootstraps_p_value=bootstraps_p_value
        )

    def plot_feature_importance(
        self, thr_pvalue: float = 1, top_n: int = None, num_cols: int = 4, save: str = None
    ):
        """
        Plot feature importance based on p-values for global and local feature importance.
        For the global feature importance, p-values are computed using ANOVA (for continuous variables)
        or Chi-Square (for categorical variables) tests. The local feature importance, p-values are measured by the
        variance and impurity of the feature within the cluster, i.e. a smaller p-value indicates lower variance/impurity.
        Feature importance is defined as log transformation of the p-value with a small offset.

        $transformed_value=-log10(p-value + \epsilon) / -log10(\epsilon)$

        where $\epsilon$ is a small positive constant (1e-50) that avoids issues with log10(0), but also significant distortion
        because $\epsilon$ is very small.
        Displays both global and local importance for top n selected features.

        :param thr_pvalue: P-value threshold for display. Only features with p-values below this threshold
                        are considered significant. Defaults to 1 (no filtering).
        :type thr_pvalue: float, optional
        :param top_n: Number of top features to display in the plot. If None, all features are included.
                    Defaults to None.
        :type top_n: int, optional
        :param num_cols: Number of plots per row in the output figure. Defaults to 4.
        :type num_cols: int, optional
        :param save: Filename to save the plot. If None, the plot will not be saved. Defaults to None.
        :type save: str, optional
        """

        # select top n features for plotting
        selected_features = self.p_value_of_features_ranked.columns.tolist()
        if top_n:
            selected_features = selected_features[:top_n]

        plotting._plot_feature_importance(
            self.p_value_of_features_ranked[selected_features],
            self.p_value_of_features_per_cluster.loc[selected_features],
            thr_pvalue,
            top_n,
            num_cols,
            save,
        )

    def plot_decision_paths(
        self,
        distributions: bool = True,
        heatmap: bool = True,
        heatmap_type: str = "static",
        thr_pvalue: float = 1,
        top_n: int = None,
        num_cols: int = 6,
        save: str = None,
    ):
        """
        Plot decision paths of the Random Forest model. This function generates visualizations
        to help understand feature importance and distribution across clusters.

        If `distributions` is `True`, it plots feature distributions per cluster using boxplots
        for continuous features and barplots for categorical features.

        If `heatmap` is `True`, it plots a heatmap of feature values sorted by clusters.

        Both plots filter and rank features based on p-values obtained from statistical tests
        (ANOVA for continuous features and chi-square for categorical features).
        In addition, all features or only the `top_n`features can be plotted. If `top_n`is `None` all features are plotted.

        :param distributions: Whether to plot feature distributions, defaults to `True`.
        :type distributions: bool, optional
        :param heatmap: Whether to plot the feature heatmap, defaults to `True`.
        :type heatmap: bool, optional
        :param thr_pvalue: P-value threshold for filtering features, defaults to `1`.
        :type thr_pvalue: float, optional
        :param top_n: Number of top features to retain after p-value ranking, defaults to `None` (no limit).
        :type top_n: int, optional
        :param num_cols: Number of plots per row in the distributions plot, defaults to `6`.
        :type num_cols: int, optional
        :param save: Filename to save the plot. If `None`, the figure is not saved, defaults to `None`.
        :type save: str, optional
        """
        # drop insignificant features
        selected_features = self.p_value_of_features_ranked.loc["p_value"] < thr_pvalue
        selected_features = self.p_value_of_features_ranked.columns[selected_features].tolist()

        # select top n features for plotting
        if top_n:
            selected_features = selected_features[:top_n]

        selected_features = ["cluster", "target"] + selected_features

        if distributions:
            plotting._plot_distributions(
                self.data_clustering_ranked[selected_features], thr_pvalue, top_n, num_cols, save
            )

        if heatmap:
            if self.model_type == "regression":
                plotting._plot_heatmap_regression(
                    self.data_clustering_ranked[selected_features],
                    thr_pvalue,
                    top_n,
                    heatmap_type,
                    save,
                )
            elif self.model_type == "classification":
                plotting._plot_heatmap_classification(
                    self.data_clustering_ranked[selected_features],
                    thr_pvalue,
                    top_n,
                    heatmap_type,
                    save,
                )



############################################
# imports
############################################


import numpy as np
import pandas as pd
from fgclustering.utils import *
from fgclustering.statistics import (
    compute_balanced_average_impurity,
    compute_total_within_cluster_variation,
    calculate_global_feature_importance,
    calculate_local_feature_importance,
)


############################################
# Tests
############################################


def test_compute_balanced_average_impurity():
    # test data
    categorical_values = pd.Series([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
    cluster_labels = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])

    result = compute_balanced_average_impurity(categorical_values, cluster_labels)

    assert result == 0.0, "error: impurity should be 0"


def test_compute_total_within_cluster_variation():
    # test data
    continuous_values = pd.Series([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
    cluster_labels = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])

    result = compute_total_within_cluster_variation(continuous_values, cluster_labels)

    assert result == 0.0, "error: within cluster variation should be 0"


def test_calculate_global_feature_importance():
    # test if anova test filters out features 1 and 2 which are the same in both clusters and
    # leaves features 3 and 4 which are clearly different in both clusters
    # parameters
    model_type = "classifier"

    # test data
    X = pd.DataFrame.from_dict(
        {
            "col_1": [1, 1, 1, 1, 1, 0.9],
            "col_2": [1, 1, 1, 1, 0.9, 0.5],
            "col_3": [1, 1, 1, 0, 0, 1],
            "col_4": [0, 0, 0, 1, 1, 1],
        }
    )
    y = pd.Series([0, 0, 0, 0, 0, 0])
    cluster_labels = np.array([0, 0, 0, 1, 1, 1])

    X_ranked, p_value_of_features = calculate_global_feature_importance(X, y, cluster_labels, model_type)

    X_ranked.drop("cluster", axis=1, inplace=True)
    assert list(X_ranked.columns) == [
        "target",
        "col_4",
        "col_3",
        "col_2",
        "col_1",
    ], "error: global feature importance returns wrong ordering"


def test_calculate_local_feature_importance():
    # test if clusterwise importance is high for feature 2 and 4 and low for feature 1 and 3
    # parameters
    thr_pvalue = 1
    bootstraps = 100
    model_type = "classifier"

    # test data
    X = pd.DataFrame.from_dict(
        {
            "col_1": [0.9, 1, 0.9, 1, 1, 0.9, 1, 0.9, 1, 0.9, 1, 0.9],
            "col_2": [0.1, 0.1, 0.1, 0.1, 0.9, 0.9, 0.9, 0.9, 1, 1, 1, 1],
            "col_3": [0.1, 0, 0.1, 0, 0.1, 0, 0.1, 0, 0.1, 0.1, 0, 0],
            "col_4": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2],
        }
    )
    y = pd.Series([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1])
    cluster_labels = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    X_ranked, p_value_of_features = calculate_global_feature_importance(X, y, cluster_labels, model_type)
    for column in X.columns:
        if p_value_of_features.loc["p_value", column] > thr_pvalue:
            X.drop(column, axis=1, inplace=True)

    p_value_of_features_per_cluster = calculate_local_feature_importance(X_ranked, bootstraps)
    importance = 1 - p_value_of_features_per_cluster
    result = importance.transpose().median()

    assert sum(result > 0.9) == 2, "error: wrong number of features with highest feature importance"



############################################
# imports
############################################

import numpy as np
import pandas as pd

from bisect import bisect
from scipy.stats import f_oneway, chisquare

from statsmodels.stats.multitest import fdrcorrection
from statsmodels.stats.proportion import proportions_chisquare
from sklearn.utils import resample

from statsmodels.stats import multitest

############################################
# functions
############################################


def compute_balanced_average_impurity(categorical_values, cluster_labels, rescaling_factor=None):
    """Compute balanced average impurity as score for categorical values in a clustering.
    Impurity score is an Gini Coefficient of the classes within each cluster.
    The class sizes are balanced by rescaling with the inverse size of the class in the overall dataset.

    :param categorical_values: Values of categorical feature / target.
    :type categorical_values: pandas.Series
    :param cluster_labels: Cluster labels for each value.
    :type cluster_labels: numpy.ndarray
    :param rescaling_factor: Dictionary with rescaling factor for each class / unique feature value.
        If parameter is set to None, the rescaling factor will be computed from the input data categorical_values, defaults to None
    :type rescaling_factor: dict
    :return: Impurity score.
    :rtype: float
    """
    # compute the number of datapoints for each class to use it then for rescaling of the
    # class sizes within each cluster --> rescaling with inverse class size
    if rescaling_factor is None:
        rescaling_factor = {
            class_: 1 / sum(categorical_values == class_) for class_ in np.unique(categorical_values)
        }
    balanced_impurities = []

    for cluster in np.unique(cluster_labels):
        categorical_values_cluster = categorical_values[cluster_labels == cluster]

        # compute balanced class probabilities (rescaled with overall class size)
        class_probabilities_unnormalized = [
            sum(categorical_values_cluster == class_) * rescaling_factor[class_]
            for class_ in np.unique(categorical_values)
        ]
        class_probabilities_unnormalized = np.array(class_probabilities_unnormalized)
        normalization_factor = class_probabilities_unnormalized.sum()
        class_probabilities = class_probabilities_unnormalized / normalization_factor

        # compute (balanced) gini impurity
        gini_impurity = 1 - np.sum(class_probabilities**2)
        balanced_impurities.append(gini_impurity)

    score = np.mean(balanced_impurities)

    return score


def compute_total_within_cluster_variation(continuous_values, cluster_labels):
    """Compute total within cluster variation as score for continuous values in a clustering.

    :param continuous_values: Values of continuous feature / target.
    :type continuous_values: pandas.Series
    :param cluster_labels: Cluster labels for each value.
    :type cluster_labels: numpy.ndarray
    :return: Within cluster variation score.
    :rtype: float
    """
    score = 0
    for cluster in np.unique(cluster_labels):
        continuous_values_cluster = continuous_values[cluster_labels == cluster]
        score += np.var(continuous_values_cluster) * len(continuous_values_cluster)

    return score


def _anova_test(list_of_df):
    """Perform one way ANOVA test on continuous features.

    :param list_of_df: List of dataframes, where each dataframe contains
        the feature values for one cluster.
    :type list_of_df: list
    :return: P-value of ANOVA test.
    :rtype: float
    """
    anova = f_oneway(*list_of_df)
    return anova.pvalue


def _chisquare_test(list_of_df):
    """Perform chi square test on categorical features.

    :param df: Dataframe with feature and cluster.
    :type df: pandas.DataFrame
    :param list_of_df: List of dataframes, where each dataframe contains
        the feature values for one cluster.
    :type list_of_df: list
    :return: P-value of chi square test.
    :rtype: float
    """
    # Retrieve how many categories exist
    # cat_vals = list({value for df_ in list_of_df for value in df_})

    # Observed counts for each cluster in each categories
    # Example for three clusters and two categories
    #            Cat 1, Cat 2
    # np.array([[5, 60],   # Cluster 1
    #           [20, 15],  # Cluster 2
    #           [25, 25]]) # Cluster 3
    # We add pseudocount 1 to avoid division by 0
    # counts_observed = np.array(
    #     [
    #         np.array([np.sum(np.array(cluster) == category) + 1 for category in cat_vals]).tolist()
    #         for cluster in list_of_df
    #     ]
    # )

    # Total counts across categories
    # counts_category_total = counts_observed.sum(axis=0)

    # Expected counts under the null hypothesis of equal preferences across categories
    # We add pseudocount 1 to avoid division by 0
    # counts_expected = (
    #     np.outer(counts_observed.sum(axis=1), counts_category_total) / counts_category_total.sum() + 1
    # )

    # Perform the chi-square test
    # _, p_value, _ = proportions_chisquare(counts_observed, counts_expected)

    # return p_value

    data = [value for df_ in list_of_df for value in df_]

    cat_vals, counts_global = np.unique(data, return_counts=True)
    count_global = counts_global / counts_global.sum()

    p_values = [
        chisquare(
            np.array([np.sum(np.array(df_) == category) for category in cat_vals]).tolist(),
            f_exp=count_global * len(df_),
        ).pvalue
        for df_ in list_of_df
    ]

    _, p_values_corrected = multitest.fdrcorrection(p_values)
    return min(p_values_corrected)


def _rank_features(data_clustering, p_value_of_features_ranked):
    """Rank features by lowest p-value.

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param y: Target column.
    :type y: pandas.Series
    :param p_value_of_features: Computed p-values of all features.
    :type p_value_of_features: dict
    :return: Ranked feature matrix.
    :rtype: pandas.DataFrame
    """
    # Reorder columns based on sorted p-values
    sorted_features = ["cluster", "target"] + p_value_of_features_ranked.columns.tolist()
    data_clustering_ranked = data_clustering[sorted_features]

    return data_clustering_ranked


def _sort_clusters_by_target(data_clustering_ranked, model_type):
    """Sort clusters by mean target values in clusters.

    :param data_clustering_ranked: Filtered and ranked data frame incl features, target and cluster numbers.
    :type data_clustering_ranked: pandas.DataFrame
    :param model_type: Model type of Random Forest model: classifier or regression.
    :type model_type: str
    :return: Filtered and ranked feature matrix with ordered clusters.
    :rtype: pandas.DataFrame
    """
    # When using a classifier, the target value is label encoded, such that we can sort the clusters by target values
    original_target = data_clustering_ranked["target"].copy()

    if model_type == "classification":
        data_clustering_ranked["target"] = data_clustering_ranked["target"].astype("category").cat.codes

    # Compute mean target values for each cluster and sort by mean values
    cluster_means = data_clustering_ranked.groupby(["cluster"])[["cluster", "target"]].mean()
    cluster_means = cluster_means.sort_values(by="target").index

    # Map the sorted clusters to a new order,  replace clusters with the new mapping and ensure the 'cluster' column is a categorical type with ordered levels
    mapping = {cluster: i + 1 for i, cluster in enumerate(cluster_means)}
    data_clustering_ranked["cluster"] = pd.Categorical(
        data_clustering_ranked["cluster"].map(mapping), ordered=True
    )

    # Restore the original target values
    data_clustering_ranked["target"] = original_target

    return data_clustering_ranked


def calculate_global_feature_importance(X, y, cluster_labels, model_type):
    """Calculate global feature importance for each feature.
    The higher the importance for a feature, the lower the p-value obtained by
    an ANOVA (continuous feature) or chi-square (categorical feature) test.
    Returned as p-value, hence importance is 1-p-value.

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param y: Target column.
    :type y: pandas.Series
    :param cluster_labels: Clustering labels.
    :type cluster_labels: numpy.ndarray
    :param model_type: Model type of Random Forest model: classifier or regression.
    :type model_type: str
    :return: Data Frame incl features, target and cluster numbers ranked by p-value of statistical test
        and dictionary with computed p-values of all features.
    :rtype: pandas.DataFrame and dict
    """
    data_clustering = pd.concat([X, y.rename("target"), pd.Series(cluster_labels, name="cluster")], axis=1)
    p_value_of_features = {}

    # statistical test for each feature
    for feature in data_clustering.columns:
        if feature not in ["cluster", "target"]:
            data_feature = data_clustering[feature]

            list_of_df = [
                data_feature[data_clustering["cluster"] == cluster].to_list()
                for cluster in data_clustering["cluster"].unique()
            ]

            # Perform statistical test based on feature type
            if isinstance(data_feature.dtype, pd.CategoricalDtype):
                p_value_of_features[feature] = _chisquare_test(list_of_df)
            elif pd.api.types.is_numeric_dtype(data_feature):
                p_value_of_features[feature] = _anova_test(list_of_df)
            else:
                raise ValueError(
                    f"Feature {feature} has dytpye {data_feature.dtype} but has to be of type category or numeric!"
                )

    # Convert p-value dictionary to a DataFrame and sort by p-value
    p_value_of_features_ranked = (
        pd.DataFrame.from_dict(p_value_of_features, orient="index", columns=["p_value"])
        .T.fillna(
            1
        )  # NaN can be produced if categorical features are dummy encoded and one feature is not present in one cluster
        .sort_values(by="p_value", axis=1)
    )

    # correct p-values for multiple testing
    _, p_values_corrected = fdrcorrection(p_value_of_features_ranked.loc["p_value"].tolist())
    p_value_of_features_ranked.loc["p_value"] = p_values_corrected

    # sort and rank clustering dataframe
    data_clustering_ranked = _rank_features(data_clustering, p_value_of_features_ranked)
    data_clustering_ranked = _sort_clusters_by_target(data_clustering_ranked, model_type)
    data_clustering_ranked.sort_values(by=["cluster", "target"], axis=0, inplace=True)

    return data_clustering_ranked, p_value_of_features_ranked


def _calculate_p_value_categorical(X_feature_cluster, X_feature, cluster, cluster_size, bootstraps):
    """Calculate bootstrapped p-value for categorical features to
    determine the importance of the feature for a certain cluster.
    The lower the bootstrapped p-value, the lower the impurity of the
    feature in the respective cluster.

    :param X_feature_cluster: Categorical feature values in cluster.
    :type X_feature_cluster: pandas.Series
    :param X_feature: Categorical feature values.
    :type X_feature: pandas.Series
    :param cluster: Cluster number.
    :type cluster: int
    :param cluster_size: Size of cluster, i.e. number of data points in cluster.
    :type cluster_size: int
    :param bootstraps: Number of bootstraps to be drawn for computation of p-value.
    :type bootstraps: int
    :return: Bootstrapped p-value for categorical feature.
    :rtype: float
    """
    cluster_label = [cluster] * cluster_size
    rescaling_factor = {class_: 1 for class_ in np.unique(X_feature)}
    X_feature_cluster_impurity = compute_balanced_average_impurity(
        X_feature_cluster, cluster_label, rescaling_factor=rescaling_factor
    )

    bootstrapped_impurity = list()
    for b in range(bootstraps):
        bootstrapped_X_feature = resample(X_feature, replace=False, n_samples=cluster_size)
        bootstrapped_impurity.append(
            compute_balanced_average_impurity(
                bootstrapped_X_feature, cluster_label, rescaling_factor=rescaling_factor
            )
        )

    bootstrapped_impurity = sorted(bootstrapped_impurity)
    p_value = bisect(bootstrapped_impurity, X_feature_cluster_impurity) / bootstraps
    return p_value


def _calculate_p_value_continuous(X_feature_cluster, X_feature, cluster_size, bootstraps):
    """Calculate bootstrapped p-value for continuous features to
    determine the importance of the feature for a certain cluster.
    The lower the bootstrapped p-value, the lower the variance of the
    feature in the respective cluster.

    :param X_feature_cluster: Continuous feature values in cluster.
    :type X_feature_cluster: pandas.Series
    :param X_feature: Continuous feature values.
    :type X_feature: pandas.Series
    :param cluster_size: Size of cluster, i.e. number of data points in cluster.
    :type cluster_size: int
    :param bootstraps: Number of bootstraps to be drawn for computation of p-value.
    :type bootstraps: int
    :return: Bootstrapped p-value for continuous feature.
    :rtype: float
    """
    X_feature_cluster_var = X_feature_cluster.var()

    bootstrapped_var = list()
    for b in range(bootstraps):
        bootstrapped_X_feature = resample(X_feature, replace=True, n_samples=cluster_size)
        bootstrapped_var.append(bootstrapped_X_feature.var())

    bootstrapped_var = sorted(bootstrapped_var)
    p_value = bisect(bootstrapped_var, X_feature_cluster_var) / bootstraps
    return p_value


def calculate_local_feature_importance(data_clustering_ranked, bootstraps_p_value):
    """Calculate local importance of each feature within each cluster.
    The higher the importance for a feature, the lower the variance (continuous feature)
    or impurity (categorical feature) of that feature within the cluster.
    Returned as p-value, hence importance is 1-p-value.

    :param data_clustering_ranked: Filtered and ranked data frame incl features, target and cluster numbers.
    :type data_clustering_ranked: pandas.DataFrame
    :param bootstraps_p_value: Number of bootstraps to be drawn for computation of p-value.
    :type bootstraps_p_value: int
    :return: p-value matrix of all features per cluster.
    :rtype: pandas.DataFrame
    """
    data_clustering_ranked = data_clustering_ranked.copy()
    clusters = data_clustering_ranked["cluster"]
    clusters_size = clusters.value_counts()
    data_clustering_ranked.drop(["cluster", "target"], axis=1, inplace=True)

    features = data_clustering_ranked.columns.tolist()
    p_value_of_features_per_cluster = pd.DataFrame(columns=clusters.unique(), index=features)

    for feature in data_clustering_ranked.columns:
        if isinstance(data_clustering_ranked[feature].dtype, pd.CategoricalDtype):
            for cluster in clusters.unique():
                X_feature_cluster = data_clustering_ranked.loc[clusters == cluster, feature]
                X_feature = data_clustering_ranked[feature]
                p_value_of_features_per_cluster.loc[feature, cluster] = _calculate_p_value_categorical(
                    X_feature_cluster,
                    X_feature,
                    cluster,
                    clusters_size.loc[cluster],
                    bootstraps_p_value,
                )

        elif pd.api.types.is_numeric_dtype(data_clustering_ranked[feature]):
            for cluster in clusters.unique():
                X_feature_cluster = data_clustering_ranked.loc[clusters == cluster, feature]
                X_feature = data_clustering_ranked[feature]
                p_value_of_features_per_cluster.loc[feature, cluster] = _calculate_p_value_continuous(
                    X_feature_cluster,
                    X_feature,
                    clusters_size.loc[cluster],
                    bootstraps_p_value,
                )

        else:
            raise ValueError(
                f"Feature {feature} has dytpye {data_clustering_ranked[feature].dtype} but has to be of type category or numeric!"
            )

    return p_value_of_features_per_cluster



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns


def plot_permutation_feature_importance(result, data, title):
    perm_sorted_idx = result.importances_mean.argsort()
    perm_indices = np.arange(0, len(result.importances_mean)) + 0.5

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4))
    fig.suptitle(title)
    ax1.barh(
        perm_indices,
        result.importances_mean[perm_sorted_idx],
        height=0.7,
        color="#3470a3",
    )
    ax1.set_yticks(perm_indices)
    ax1.set_yticklabels(data.columns[perm_sorted_idx])
    ax1.set_ylim((0, len(result.importances_mean)))
    ax2.boxplot(
        result.importances[perm_sorted_idx].T,
        vert=False,
        labels=data.columns[perm_sorted_idx],
    )
    fig.tight_layout()
    plt.show()


def plot_impurity_feature_importance(importance, names, title):
    # Create arrays from feature importance and feature names
    feature_importance = np.array(importance)
    feature_names = np.array(names)

    # Create a DataFrame using a Dictionary
    data = {"feature_names": feature_names, "feature_importance": feature_importance}
    fi_df = pd.DataFrame(data)

    # Sort the DataFrame in order decreasing feature importance
    fi_df.sort_values(by=["feature_importance"], ascending=False, inplace=True)

    # Define size of bar plot
    plt.figure(figsize=(5, 4))
    # Plot Searborn bar chart
    sns.barplot(
        x=fi_df["feature_importance"], y=fi_df["feature_names"], color="#3470a3"
    )
    # Add chart labels
    plt.title(title)
    plt.xlabel("feature importance")
    plt.ylabel("feature names")


def plot_correlation_matrix(data):
    f, ax = plt.subplots(figsize=(5, 5))
    corr = data.corr()
    mask = np.triu(np.ones_like(corr, dtype=bool))
    np.fill_diagonal(mask, False)
    sns.heatmap(
        round(corr, 2),
        mask=mask,
        cmap=sns.diverging_palette(220, 10, as_cmap=True),
        square=True,
        ax=ax,
        annot=True,
    )



############################################
# imports
############################################

import numpy as np
import pandas as pd

import kmedoids
from sklearn.datasets import make_classification, make_regression
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from fgclustering.utils import proximityMatrix
from fgclustering.optimizer import (
    optimizeK,
    _compute_stability_indices_parallel,
    _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster,
)


############################################
# Tests
############################################


def test_optimizeK():
    # parameters
    method_clustering = "pam"
    init_clustering = "random"
    max_iter_clustering = 100
    discart_value_JI = 0.6
    bootstraps_JI = 100
    random_state = 42
    n_jobs = 3
    verbose = 1

    ### test classification
    max_K = 5
    model_type = "classification"

    # test data
    X, y = make_classification(
        n_samples=300,
        n_features=10,
        n_informative=4,
        n_redundant=2,
        n_classes=2,
        n_clusters_per_class=1,
        random_state=1,
    )
    X = pd.DataFrame(X)

    model = RandomForestClassifier(
        max_depth=10,
        max_features="sqrt",
        max_samples=0.8,
        bootstrap=True,
        oob_score=True,
        random_state=42,
    )
    model.fit(X, y)

    distance_matrix = 1 - proximityMatrix(model, X)
    k = optimizeK(
        distance_matrix,
        y,
        model_type,
        max_K,
        method_clustering,
        init_clustering,
        max_iter_clustering,
        discart_value_JI,
        bootstraps_JI,
        random_state,
        n_jobs,
        verbose,
    )

    assert k == 2, "Error optimal number of clusters for classification problem is not equal to 2"

    ### test regression
    max_K = 7
    model_type = "regression"
    discart_value_JI = 0.7

    # test data
    X, y = make_regression(
        n_samples=500,
        n_features=10,
        n_informative=4,
        n_targets=1,
        noise=0,
        random_state=1,
    )
    X = pd.DataFrame(X)

    model = RandomForestRegressor(
        max_depth=5,
        max_features="sqrt",
        max_samples=0.8,
        bootstrap=True,
        oob_score=True,
        random_state=42,
    )
    model.fit(X, y)

    distance_matrix = 1 - proximityMatrix(model, X)
    k = optimizeK(
        distance_matrix,
        y,
        model_type,
        max_K,
        method_clustering,
        init_clustering,
        max_iter_clustering,
        discart_value_JI,
        bootstraps_JI,
        random_state,
        n_jobs,
        verbose,
    )

    assert (
        k == 2 or k == 3 or k == 4
    ), "Error optimal number of clusters for regression problem is not equal to 2 or 3 or 4"


def test_compute_stability_indices():
    # parameters
    method_clustering = "pam"
    init_clustering = "random"
    max_iter_clustering = 100
    bootstraps_JI = 100
    random_state = 42
    n_jobs = 3

    # test data
    distance_matrix = 1 - np.kron(np.eye(3, dtype=int), np.ones([10, 10]))

    # test 1: test if 3 different clusters are found and have maximal stability
    cluster_method = (
        lambda X: kmedoids.KMedoids(
            n_clusters=3,
            method=method_clustering,
            init=init_clustering,
            metric="precomputed",
            max_iter=max_iter_clustering,
            random_state=random_state,
        )
        .fit(X)
        .labels_
    )
    labels = cluster_method(distance_matrix)

    result = _compute_stability_indices_parallel(
        distance_matrix, labels, cluster_method, bootstraps_JI, n_jobs
    )
    assert result[0] == 1.0, "Clusters that should be stable are found to be unstable"
    assert result[1] == 1.0, "Clusters that should be stable are found to be unstable"
    assert result[2] == 1.0, "Clusters that should be stable are found to be unstable"

    # test 2: test if 2 different clusters are found that don't have maximal stability
    cluster_method = (
        lambda X: kmedoids.KMedoids(
            n_clusters=2,
            method=method_clustering,
            init=init_clustering,
            metric="precomputed",
            max_iter=max_iter_clustering,
            random_state=random_state,
        )
        .fit(X)
        .labels_
    )
    labels = cluster_method(distance_matrix)
    result = _compute_stability_indices_parallel(
        distance_matrix, labels, cluster_method, bootstraps_JI, n_jobs
    )

    assert min(result[0], result[1]) < 1.0, "Clusters that should be unstable are found to be stable"


def test_translate_cluster_labels_to_dictionary_of_index_sets_per_cluster():
    # test data
    labels = [1, 2, 3, 2, 1]
    mapping = {0: 10, 1: 11, 2: 12, 3: 13, 4: 14, 5: 15}

    # test translation without mapping
    dictionary_of_index_sets_per_cluster = _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(
        labels, mapping=False
    )
    assert dictionary_of_index_sets_per_cluster == {
        1: set([0, 4]),
        2: set([1, 3]),
        3: set([2]),
    }, "error: wrong dictionary of index sets"

    # test translation with mapping
    dictionary_of_index_sets_per_cluster = _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(
        labels, mapping=mapping
    )
    assert dictionary_of_index_sets_per_cluster == {
        1: set([10, 14]),
        2: set([11, 13]),
        3: set([12]),
    }, "error: wrong dictionary of index sets"
